import numpy as np
import scipy.stats


class MixedSplitter(object):
    def __init__(self,
                 x,
                 y,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 class_targets=None):
        self.n_train = x.shape[0]
        self.n_features = x.shape[1]
        self.n_targets = y.shape[1]
        self.class_targets = class_targets if class_targets else []
        self.max_features = max_features
        self.min_samples_leaf = min_samples_leaf
        self.root_impurity = self.__impurity_node(y)
        self.choose_split = choose_split

    def split(self, x, y):
        if self.max_features == 'sqrt':
            self.max_features = int(np.ceil(np.sqrt(self.n_features)))

        return self.__find_best_split(x, y)

    def __find_best_split(self, x, y):
        if x.shape[0] <= self.min_samples_leaf:
            return None, None, np.inf

        best_f = None
        best_v = None
        best_imp = -np.inf

        try_features = np.random.choice(
            np.arange(self.n_features),
            self.max_features,
            replace=False
        )

        for f in try_features:
            values = np.unique(x[:, f])
            if values.size < 2:
                continue

            values = (values[:-1] + values[1:]) / 2

            #random value subsampling
            values = np.random.choice(values, min(2, values.size))
            for v in values:
                imp = self.__try_split(x, y, f, v)
                if imp > best_imp:
                    best_f, best_v, best_imp = f, v, imp

        return best_f, best_v, best_imp

    def __try_split(self, x, y, f, t):
        l_idx = x[:, f] <= t
        r_idx = x[:, f] > t

        return self.__impurity_split(y, y[l_idx, :], y[r_idx, :])

    def __impurity_node(self, y):
        def impurity_class(y):
            y = y.astype(int)
            freq = np.bincount(y) / y.size
            freq = freq[freq != 0]
            return 0 - np.array([f * np.log2(f) for f in freq]).sum()

        def impurity_reg(y):
            if np.unique(y).size < 2:
                return 0

            n_bins = 100
            freq, _ = np.histogram(y, bins=n_bins, density=True)
            proba = (freq + 1) / (freq.sum() + n_bins)
            bin_width = (y.max() - y.min()) / n_bins

            return 0 - bin_width * (proba * np.log2(proba)).sum()

        delta = 0.0001
        imp = np.zeros(self.n_targets)
        for i in range(self.n_targets):
            if i in self.class_targets:
                imp[i] = impurity_class(y[:, i]) + delta
            else:
                imp[i] = impurity_reg(y[:, i]) + delta
        return imp

    def __impurity_split(self, y, y_left, y_right):
        n_parent = y.shape[0]
        n_left = y_left.shape[0]
        n_right = y_right.shape[0]

        if n_left < self.min_samples_leaf or n_right < self.min_samples_leaf:
            return np.inf
        else:
            imp_left = self.__impurity_node(y_left) / self.root_impurity
            imp_right = self.__impurity_node(y_right) / self.root_impurity
            imp_parent = self.__impurity_node(y) / self.root_impurity

            gain_left = (n_left / n_parent) * (imp_parent - imp_left)
            gain_right = (n_right / n_parent) * (imp_parent - imp_right)
            gain = gain_left + gain_right

            #imp = (imp_left + imp_right) / imp_parent

            if self.choose_split == 'mean':
                return gain.mean()
            else:
                return gain.max()


class MixedRandomTree(object):
    def __init__(self,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 class_targets=None):
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.class_targets = class_targets if class_targets else []
        self.choose_split = choose_split

    def fit(self, x, y):
        if y.ndim == 1:
            y = y.reshape((y.size, 1))

        self.n_targets = y.shape[1]

        self.splitter = MixedSplitter(x,
                                      y,
                                      self.max_features,
                                      self.min_samples_leaf,
                                      self.choose_split,
                                      self.class_targets)

        split_f = []
        split_t = []
        leaf_value = []
        l_child = []
        r_child = []
        n_i = []

        split_queue = [(x, y)]
        i = 0
        while len(split_queue) > 0:
            next_x, next_y = split_queue.pop(0)

            leaf_value.append(self._make_leaf(next_y))
            n_i.append(next_y.shape[0])

            f, t, imp = self.splitter.split(next_x, next_y)

            split_f.append(f)
            split_t.append(t)
            if f:
                l_child.append(i + len(split_queue) + 1)
                r_child.append(i + len(split_queue) + 2)
            else:
                l_child.append(None)
                r_child.append(None)

            if f:
                l_idx = next_x[:, f] <= t
                r_idx = next_x[:, f] > t

                split_queue.append((next_x[l_idx, :], next_y[l_idx, :]))
                split_queue.append((next_x[r_idx, :], next_y[r_idx, :]))

            i += 1

        self.f = np.array(split_f)
        self.t = np.array(split_t)
        self.v = np.array(leaf_value)
        self.l = np.array(l_child)
        self.r = np.array(r_child)
        self.n = np.array(n_i)

    def _make_leaf(self, y):
        y_ = np.zeros(self.n_targets)
        for i in range(self.n_targets):
            if i in self.class_targets:
                y_[i] = np.argmax(np.bincount(y[:, i].astype(int)))
            else:
                y_[i] = y[:, i].mean()
        return y_

    def predict(self, x):
        n_test = x.shape[0]
        pred = np.zeros((n_test, self.n_targets))

        def traverse(x, test_idx, node_idx):
            if test_idx.size < 1:
                return

            if not self.f[node_idx]:
                pred[test_idx, :] = self.v[node_idx]
            else:
                l_idx = x[:, self.f[node_idx]] <= self.t[node_idx]
                r_idx = x[:, self.f[node_idx]] > self.t[node_idx]

                traverse(x[l_idx, :], test_idx[l_idx], self.l[node_idx])
                traverse(x[r_idx, :], test_idx[r_idx], self.r[node_idx])

        traverse(x, np.arange(n_test), 0)
        return pred

    def print(self):
        def print_l(level, i):
            if self.f[i]:
                print('\t' * level + '[{} <= {}]:'.format(self.f[i], self.t[i]))
                print_l(level + 1, self.l[i])
                print_l(level + 1, self.r[i])
            else:
                print('\t' * level + str(self.v[i]) + ' ({})'.format(self.n[i]))

        print_l(0, 0)

class MixedRandomForestLegacy(object):
    def __init__(self,
                 n_estimators=10,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 class_targets=None):
        self.n_estimators = n_estimators
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.class_targets = class_targets if class_targets else []
        self.choose_split = choose_split

    def fit(self, x, y):
        self.estimators = []

        if y.ndim == 1:
            y = y.reshape((y.size, 1))
        self.n_targets = y.shape[1]

        self.class_labels = {}
        for i in filter(lambda x: x in self.class_targets, range(self.n_targets)):
            self.class_labels[i] = np.unique(y[:, i])

        n_train = x.shape[0]
        for i in range(self.n_estimators):
            m = MixedRandomTree(self.max_features,
                                self.min_samples_leaf,
                                self.choose_split,
                                self.class_targets)
            sample_idx = np.random.choice(np.arange(n_train),
                                          n_train,
                                          replace=True)

            m.fit(x[sample_idx, :], y[sample_idx, :])
            self.estimators.append(m)

    def predict(self, x):
        n_test = x.shape[0]
        pred = np.zeros((n_test, self.n_targets, self.n_estimators))
        for i, m in enumerate(self.estimators):
            pred[:, :, i] = m.predict(x)

        pred_avg = np.zeros((n_test, self.n_targets))
        for i in range(self.n_targets):
            if i in self.class_targets:
                pred_avg[:, i], _ = scipy.stats.mode(pred[:, i, :].T)
            #pred_avg[:, i] = np.argmax(np.bincount(pred[:, i, :]))
            else:
                pred_avg[:, i] = pred[:, i, :].mean(axis=1)

        return pred_avg

    def predict_proba(self, x):
        n_test = x.shape[0]
        pred = np.zeros((n_test, self.n_targets, self.n_estimators))
        for i, m in enumerate(self.estimators):
            pred[:, :, i] = m.predict(x)

        pred_avg = np.zeros((n_test, self.n_targets), dtype=object)
        for i in range(self.n_targets):
            if i in self.class_targets:
                for j in range(n_test):
                    freq = np.bincount(pred[j, i, :].T.astype(int),
                                       minlength=self.class_labels[i].size)
                    pred_avg[j, i] = freq / self.n_estimators
            else:
                pred_avg[:, i] = pred[:, i, :].mean(axis=1)

        return pred_avg


def acc(y, y_hat):
    return (y.astype(int) == y_hat.astype(int)).sum() / y.size

def rmse(y, y_hat):
    return np.sqrt(((y - y_hat) ** 2).mean())

def cross_validation(model,
                     x,
                     y,
                     folds=10,
                     class_targets=None,
                     class_eval=acc,
                     reg_eval=rmse,
                     verbose=False):
    import copy

    class_targets = class_targets if class_targets else []

    idx = np.random.permutation(x.shape[0])
    fold_size = int(idx.size / folds)

    if y.ndim == 1:
        y = y.reshape((y.size, 1))

    y_hat = np.zeros((idx.size, y.shape[1]))

    for i in range(folds):
        if verbose:
            print('Running fold {} of {} ...'.format(i + 1, folds))

        fold_start = i * fold_size
        fold_stop = min((i + 1) * fold_size, idx.size)

        mask = np.ones(idx.size, dtype=bool)
        mask[fold_start:fold_stop] = 0

        train_idx = idx[mask]
        test_idx = idx[(1 - mask).astype(bool)]

        m = copy.copy(model)
        m.fit(x[train_idx, :], y[train_idx, :])
        y_hat[test_idx, :] = m.predict(x[test_idx, :])

    scores = np.zeros(y.shape[1])
    for i in range(y.shape[1]):
        if i in class_targets:
            scores[i] = class_eval(y[:, i], y_hat[:, i])
        else:
            scores[i] = reg_eval(y[:, i], y_hat[:, i])

    return scores


import numpy as np
from numba import njit


@njit
def get_bin_edges(a, bins):
    bin_edges = np.zeros((bins + 1,), dtype=np.float64)
    a_min = a.min()
    a_max = a.max()
    delta = (a_max - a_min) / bins
    for i in range(bin_edges.shape[0]):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid round-off error on last point
    return bin_edges


@njit
def compute_bin(x, bin_edges):
    # assuming uniform bins for now
    n = bin_edges.shape[0] - 1
    a_min = bin_edges[0]
    a_max = bin_edges[-1]

    # special case to mirror NumPy behavior for last bin
    if x == a_max:
        return n - 1  # a_max always in last bin

    bin = int(n * (x - a_min) / (a_max - a_min))

    if bin < 0 or bin >= n:
        return None
    else:
        return bin


@njit
def numba_histogram(a, bins):
    hist = np.zeros((bins,), dtype=np.intp)
    bin_edges = get_bin_edges(a, bins)

    for x in a.flat:
        bin = compute_bin(x, bin_edges)
        if bin is not None:
            hist[int(bin)] += 1

    return hist, bin_edges



import numpy as np
from numba import njit

from morfist.algo.histogram import numba_histogram


@njit
def impurity_classification(y_classification):
    # Calculate the impurity value for the classification task

    # Cast to integer
    y_class = y_classification.astype(np.int8)

    # Calculate frequencies
    frequency = np.bincount(y_class) / y_class.size

    result = 0
    for i in range(frequency.size):
        if frequency[i]:
            result += frequency[i] * np.log2(frequency[i])

    return 0 - result


@njit
def impurity_regression(y, y_regression):
    # Calculate the impurity value for the regression task

    if np.unique(y_regression).size < 2:
        return 0

    n_bins = 100
    bin_width = (y.max() - y.min()) / n_bins

    frequency, _ = numba_histogram(y_regression, n_bins)
    frequency_float = frequency.astype(np.float64)
    frequency_float = (frequency_float / len(y)) / bin_width

    probability = (frequency_float + 1) / (frequency_float.sum() + n_bins)

    return 0 - bin_width * (probability * np.log2(probability)).sum()


@njit
def get_gain(imp_n_left, imp_n_right, imp_n, imp_root, n_left, n_right, n_parent):
    impurity_left = imp_n_left / imp_root
    impurity_right = imp_n_right / imp_root
    impurity_parent = imp_n / imp_root

    gain_left = (n_left / n_parent) * (impurity_parent - impurity_left)
    gain_right = (n_right / n_parent) * (impurity_parent - impurity_right)
    return gain_left + gain_right


def get_max_features(max_features, n_features):
    # Maximum number of features to try for the best split
    n_max_features = n_features
    if max_features == 'sqrt':
        n_max_features = int(np.sqrt(n_features))
    elif max_features == 'log2':
        n_max_features = int(np.log2(n_features))
    elif isinstance(max_features, float):
        n_max_features = int(max_features * n_features)
    elif max_features is None:
        n_max_features = n_features

    return n_max_features


class MixedSplitter:
    def __init__(self,
                 x,
                 y,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 classification_targets=None):
        """Class in charge of finding the best split at every given moment

        :param x: training data
        :param y: target data
        :param max_features: the number of features to consider when looking for the best split
        :param min_samples_leaf: minimum amount of samples in each leaf
        :param choose_split:  method used to find the best split
        :param classification_targets: features that are part of the classification task
        """
        self.n_train = x.shape[0]
        self.n_features = x.shape[1]
        self.n_targets = y.shape[1]
        self.classification_targets = classification_targets if classification_targets is not None else []
        self.max_features = get_max_features(max_features, self.n_features)
        self.min_samples_leaf = min_samples_leaf
        self.root_impurity = self.__impurity_node(y)
        self.choose_split = choose_split

    def split(self, x, y):
        # If there are not enough features in the leaf, stop splitting
        if x.shape[0] <= self.min_samples_leaf:
            return None, None, np.inf

        # Best feature
        best_feature = None
        # Best value
        best_value = None
        # Best impurity
        best_impurity = -np.inf

        # Random selection of the features to try for the best split
        try_features = np.random.choice(
            np.arange(self.n_features),
            self.max_features,
            replace=False
        )

        # Try each of the selected features and find which of them gives the best split(higher impurity)
        for feature in try_features:
            # Get the unique possible values for this particular feature
            values = np.unique(x[:, feature])

            # Split value selection(random value subsampling): Boström (2011)
            #   Two random feature values are selected, and a split is attempted at their mean
            if values.size < 2:
                continue
            values = (values[:-1] + values[1:]) / 2
            value = np.random.choice(values)

            # Try to split with this specific combination of feature and value
            left_idx = x[:, feature] <= value
            right_idx = x[:, feature] > value

            impurity = self.__impurity_split(y, y[left_idx, :], y[right_idx, :])
            # If it's better than the previous saved one, save the values
            if impurity > best_impurity:
                best_feature, best_value, best_impurity = feature, value, impurity

        return best_feature, best_value, best_impurity

    # Calculate the impurity of a split
    def __impurity_split(self, y_parent, y_left, y_right):
        n_left = y_left.shape[0]
        n_right = y_right.shape[0]
        if n_left < self.min_samples_leaf or n_right < self.min_samples_leaf:
            return np.inf
        else:
            n_parent = y_parent.shape[0]

            gain = get_gain(self.__impurity_node(y_left),
                            self.__impurity_node(y_right),
                            self.__impurity_node(y_parent),
                            self.root_impurity,
                            n_left,
                            n_right,
                            n_parent)

            if self.choose_split == 'mean':
                return gain.mean()
            elif self.choose_split == 'random':
                return np.random.choice(gain)
            else:
                return gain.max()

    def __impurity_node(self, y):
        # Calculate the impurity of a node
        delta = 0.0001
        impurity = np.zeros(self.n_targets)
        # Calculate the impurity value for each of the targets(classification or regression)
        for i in range(self.n_targets):
            if i in self.classification_targets:
                impurity[i] = impurity_classification(y[:, i]) + delta
            else:
                impurity[i] = impurity_regression(y, y[:, i]) + delta
        return impurity



import copy

import numpy as np


def accuracy(y, y_hat):
    # Calculate classification accuracy of model
    return (y.astype(int) == y_hat.astype(int)).sum() / y.size


def rmse(y, y_hat):
    # Calculate root squared mean error of model
    return np.sqrt(((y - y_hat) ** 2).mean())


def cross_validation(model,
                     x,
                     y,
                     folds=10,
                     classification_targets=None,
                     classification_eval=accuracy,
                     reg_eval=rmse,
                     verbose=False):
    """Perform cross validation on a model

    :param model:  model to be validated
    :param x: X values of the data set
    :param y: Y values of the data set
    :param folds: number of folds
    :param classification_targets: features that are part of the classification task
    :param classification_eval: function to evaluate model classification accuracy
    :param reg_eval: function to evaluate model regression accuracy
    :param verbose: used for debug purposes
    :return: scores[]:
                 0: classification accuracy
                 1: regression RMSE
    """
    classification_targets = classification_targets if classification_targets is not None else []

    idx = np.random.permutation(x.shape[0])
    fold_size = int(idx.size / folds)

    if y.ndim == 1:
        y = y.reshape((y.size, 1))

    y_hat = np.zeros((idx.size, y.shape[1]))

    # Perform the cross-validation
    # Train and fit the model for different subsets of the data
    for i in range(folds):
        if verbose:
            print('Running fold {} of {} ...'.format(i + 1, folds))

        fold_start = i * fold_size
        fold_stop = min((i + 1) * fold_size, idx.size)

        mask = np.ones(idx.size, dtype=bool)
        mask[fold_start:fold_stop] = 0

        train_idx = idx[mask]
        test_idx = idx[(1 - mask).astype(bool)]

        m = copy.copy(model)
        m.fit(x[train_idx, :], y[train_idx, :])
        y_hat[test_idx, :] = m.predict(x[test_idx, :])

    scores = np.zeros(y.shape[1])
    # Calculate the classification and regression accuracy of the model
    for i in range(y.shape[1]):
        if i in classification_targets:
            scores[i] = classification_eval(y[:, i], y_hat[:, i])
        else:
            scores[i] = reg_eval(y[:, i], y_hat[:, i])

    return scores



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['morfist', 'morfist.algo', 'morfist.core', 'morfist.legacy']

package_data = \
{'': ['*']}

install_requires = \
['numba>=0.51.0,<0.52.0', 'numpy>=1.19.1,<2.0.0', 'scipy>=1.5.2,<2.0.0']

setup_kwargs = {
    'name': 'decision-tree-morfist',
    'version': '0.3.3',
    'description': 'Multi-target Random Forest implementation that can mix both classification and regression tasks.',
    'long_description': "# morfist: mixed-output-rf\nMulti-target Random Forest implementation that can mix both classification and regression tasks.\n\nMorfist implements the Random Forest algorithm (Breiman, 2001)\nwith support for mixed-task multi-task learning, i.e., it is possible to train the model on any number\nof classification tasks and regression tasks, simultaneously. Morfist's mixed multi-task learning implementation follows that proposed by Linusson (2013). \n\n* [Breiman, L. (2001). Random forests. Machine learning, 45(1), 5-32](https://link.springer.com/article/10.1023%2FA%3A1010933404324).\n* [Linusson, H. (2013). Multi-output random forests](https://pdfs.semanticscholar.org/4219/f87ed41c558d43cf78f63976cf87bcd7ebb0.pdf).\n\n## Installation\n\nWith pip:\n```\npip install decision-tree-morfist\n```\nWith conda:\n```\nconda install -c systemallica decision-tree-morfist\n```\n## Usage\n\n### Initialising the model\n\n- Similarly to a scikit-learn [RandomForestClassifier](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html), a MixedRandomForest can be initialised in this way:\n```\nfrom morfist import MixedRandomForest\n\nmrf = MixedRandomForest(\n    n_estimators=n_trees,\n    min_samples_leaf=1,\n    classification_targets=[0]\n)\n```\n\nFor more info on the possible parameters, visit the [documentation](https://systemallica.github.io/morfist/).\n\n### Training the model\n\n- Once the model is initialised, it can be fitted like this:\n    ```\n    mrf.fit(X, y)\n    ```\n    Where X are the training examples and Y are their respective labels(if they are categorical) or values(if they are numerical)\n\n### Prediction\n\n- The model can be now used to predict new instances.\n    - Class/value:\n    ```\n    mrf.predict(x)\n    ```\n    - Probability:\n    ```\n    mrf.predict_proba(x)\n    ```\n  \n## TODO:\n* Speed up the learning algorithm implementation (morfist is currently **much** slower than the Random Forest implementation available in scikit-learn) \n",
    'author': 'Andrés Reverón Molina',
    'author_email': 'andres@reveronmolina.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/systemallica/morfist',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



import numpy as np

from morfist.core.MixedSplitter import MixedSplitter


class MixedRandomTree:
    def __init__(self,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 classification_targets=None):
        """Build a Random Tree

        :param max_features: the number of features to consider when looking for the best split
        :param min_samples_leaf: minimum amount of samples in each leaf
        :param choose_split: method used to find the best split
        :param classification_targets: features that are part of the classification task
        """
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.classification_targets = classification_targets if classification_targets is not None else []
        self.choose_split = choose_split
        self.n_targets = 0
        self.features = []
        self.values = []
        self.leaf_values = []
        self.left_children = []
        self.right_children = []
        self.n = []

    def fit(self, x, y):
        if y.ndim == 1:
            y = y.reshape((y.size, 1))

        self.n_targets = y.shape[1]

        splitter = MixedSplitter(x,
                                 y,
                                 self.max_features,
                                 self.min_samples_leaf,
                                 self.choose_split,
                                 self.classification_targets)

        split_features = []
        split_values = []
        leaf_values = []
        left_children = []
        right_children = []
        n_i = []

        split_queue = [(x, y)]
        i = 0
        # Build the tree until all values are covered
        while len(split_queue) > 0:
            next_x, next_y = split_queue.pop(0)

            leaf_values.append(self._make_leaf(next_y))
            n_i.append(next_y.shape[0])

            feature, value, impurity = splitter.split(next_x, next_y)

            split_features.append(feature)
            split_values.append(value)
            if feature:
                left_children.append(i + len(split_queue) + 1)
                right_children.append(i + len(split_queue) + 2)

                l_idx = next_x[:, feature] <= value
                r_idx = next_x[:, feature] > value

                split_queue.append((next_x[l_idx, :], next_y[l_idx, :]))
                split_queue.append((next_x[r_idx, :], next_y[r_idx, :]))
            else:
                left_children.append(None)
                right_children.append(None)

            i += 1

        self.features = np.array(split_features)
        self.values = np.array(split_values)
        self.leaf_values = np.array(leaf_values)
        self.left_children = np.array(left_children)
        self.right_children = np.array(right_children)
        self.n = np.array(n_i)

    def _make_leaf(self, y):
        y_ = np.zeros(self.n_targets)
        for i in range(self.n_targets):
            if i in self.classification_targets:
                y_[i] = np.argmax(np.bincount(y[:, i].astype(int)))
            else:
                y_[i] = y[:, i].mean()
        return y_

    def predict(self, x):
        n_test = x.shape[0]
        prediction = np.zeros((n_test, self.n_targets))

        def traverse(x_traverse, test_idx, node_idx):
            if test_idx.size < 1:
                return

            if not self.features[node_idx]:
                prediction[test_idx, :] = self.leaf_values[node_idx]
            else:
                left_idx = x_traverse[:, self.features[node_idx]] <= self.values[node_idx]
                right_idx = x_traverse[:, self.features[node_idx]] > self.values[node_idx]

                traverse(x_traverse[left_idx, :], test_idx[left_idx], self.left_children[node_idx])
                traverse(x_traverse[right_idx, :], test_idx[right_idx], self.right_children[node_idx])

        traverse(x, np.arange(n_test), 0)
        return prediction

    def print(self):
        def print_level(level, i):
            if self.features[i]:
                print('\t' * level + '[{} <= {}]:'.format(self.features[i], self.values[i]))
                print_level(level + 1, self.left_children[i])
                print_level(level + 1, self.right_children[i])
            else:
                print('\t' * level + str(self.leaf_values[i]) + ' ({})'.format(self.n[i]))

        print_level(0, 0)



import numpy as np
import scipy.stats

from morfist.core.MixedRandomTree import MixedRandomTree


class MixedRandomForest:
    def __str__(self):
        params = "("
        i = 0
        for key in self.__dict__:
            if i == 0:
                params += str(key) + "=" + str(self.__dict__[key]) + ", \n"
            elif i == len(self.__dict__) - 1:
                params += "\t\t\t" + str(key) + "=" + str(self.__dict__[key]) + ")"
            else:
                params += "\t\t\t" + str(key) + "=" + str(self.__dict__[key]) + ", \n"
            i += 1
        return self.__class__.__name__ + params

    def __init__(self,
                 n_estimators=10,
                 max_features='sqrt',
                 min_samples_leaf=5,
                 choose_split='mean',
                 classification_targets=None):
        """Build a printable Random Forest model

        :param n_estimators: number of trees in the forest
        :param max_features: the number of features to consider when looking for the best split
        :param min_samples_leaf: minimum amount of samples in each leaf
        :param choose_split: method to use to find the best split
        :param classification_targets: features that are part of the classification task
        """
        self.n_estimators = n_estimators
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.classification_targets = classification_targets if classification_targets is not None else []
        self.choose_split = choose_split
        self.n_targets = 0
        self.classification_labels = {}
        self.estimators = []

    # Fit the model
    def fit(self, x, y):
        self.estimators = []

        if y.ndim == 1:
            y = y.reshape((y.size, 1))
        self.n_targets = y.shape[1]

        # Get the classification labels
        # It takes the unique labels of the specified classification variables
        for i in filter(lambda j: j in self.classification_targets, range(self.n_targets)):
            self.classification_labels[i] = np.unique(y[:, i])

        n_train = x.shape[0]
        # Train the random trees that are part of the forest
        for i in range(self.n_estimators):
            m = MixedRandomTree(self.max_features,
                                self.min_samples_leaf,
                                self.choose_split,
                                self.classification_targets)

            # It is a random forest so the trees are built with random subsets of the data
            sample_idx = np.random.choice(np.arange(n_train),
                                          n_train,
                                          replace=True)

            m.fit(x[sample_idx, :], y[sample_idx, :])
            self.estimators.append(m)

    # Predict the class/value of an instance
    def predict(self, x):
        n_test = x.shape[0]
        pred = np.zeros((n_test, self.n_targets, self.n_estimators))
        for i, m in enumerate(self.estimators):
            pred[:, :, i] = m.predict(x)

        pred_avg = np.zeros((n_test, self.n_targets))
        for i in range(self.n_targets):
            # Predict categorical value
            if i in self.classification_targets:
                pred_avg[:, i], _ = scipy.stats.mode(pred[:, i, :].T)
            # Predict numerical value
            else:
                pred_avg[:, i] = pred[:, i, :].mean(axis=1)

        return pred_avg

    # Predict the probability of an instance
    def predict_proba(self, x):
        n_test = x.shape[0]
        pred = np.zeros((n_test, self.n_targets, self.n_estimators))
        for i, m in enumerate(self.estimators):
            pred[:, :, i] = m.predict(x)

        pred_avg = np.zeros((n_test, self.n_targets), dtype=object)
        for i in range(self.n_targets):
            if i in self.classification_targets:
                for j in range(n_test):
                    freq = np.bincount(pred[j, i, :].T.astype(int),
                                       minlength=self.classification_labels[i].size)

                    # Fix for a weird bug TODO: find the root cause
                    if len(freq) > self.classification_labels[i].size:
                        freq_fixed = np.zeros(self.classification_labels[i].size)
                        for x in range(len(self.classification_labels[i])):
                            idx = int(self.classification_labels[i][x])
                            freq_fixed[x] = freq[idx]
                        pred_avg[j, i] = freq_fixed / self.n_estimators
                    else:
                        pred_avg[j, i] = freq / self.n_estimators
            else:
                pred_avg[:, i] = pred[:, i, :].mean(axis=1)

        return pred_avg
